import argparse

import requests

import logging

logger = logging.getLogger("pynxtools")

rod_id = 1000679


def save_rod_file_from_ROD_via_API(rod_id: int):
    url = "https://solsa.crystallography.net/rod/" + str(rod_id) + ".rod"

    logger.info(f"Initialized download of .rod file with ID '{rod_id}' from '{url}'.")

    try:
        response = requests.post(url)
        response.raise_for_status()  # Raise HTTP error for bad

        logger.info(f"Successfully received .rod file with ID '{rod_id}'")

        filename = str(rod_id)

        with open(filename + ".rod", "w", encoding="utf-8") as file:
            file.write(response.text)
        logger.info(f"Saved .rod file with ID '{rod_id}' to file '{filename}'")

    except requests.exceptions.ConnectionError as con_err:
        logger.error(f"ConnectionError occured: {con_err}")
    except requests.exceptions.HTTPError as http_err:
        logger.error(f"CHTTPError occured: {http_err}")
    except requests.exceptions.RequestException as req_exc:
        logger.error(f"RequestException occured: {req_exc}")


def trigger_rod_download():
    # Create an argument parser
    parser = argparse.ArgumentParser(description="Download a CIF file.")
    parser.add_argument(
        "rod_id",  # The argument's name
        type=str,  # Argument type (e.g., string)
        help="The name of the file to download",  # Help message
    )

    # Parse the command-line arguments
    args = parser.parse_args()

    save_rod_file_from_ROD_via_API(args.rod_id)



# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""An example reader implementation based on the MultiFormatReader."""

import copy
import datetime
import logging
import re
from pathlib import Path
from typing import Any, Dict, List, Tuple  # Optional, Union, Set

from pynxtools.dataconverter.readers.multi.reader import MultiFormatReader
from pynxtools.dataconverter.readers.utils import parse_yml

from pynxtools_raman.rod.rod_reader import RodParser, post_process_rod
from pynxtools_raman.witec.witec_reader import parse_txt_file, post_process_witec

logger = logging.getLogger("pynxtools")

CONVERT_DICT: Dict[str, str] = {}

REPLACE_NESTED: Dict[str, str] = {}


class RamanReader(MultiFormatReader):
    """MyDataReader implementation for the DataConverter to convert mydata to NeXus."""

    supported_nxdls = ["NXraman"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.raman_data_dicts: List[Dict[str, Any]] = []
        self.raman_data: Dict[str, Any] = {}
        self.eln_data: Dict[str, Any] = {}
        self.config_file: Path

        self.missing_meta_data = None

        self.extensions = {
            ".yml": self.handle_eln_file,
            ".yaml": self.handle_eln_file,
            ".txt": self.handle_txt_file,
            ".json": self.set_config_file,
            ".rod": self.handle_rod_file,
        }

    def set_config_file(self, file_path: Path) -> Dict[str, Any]:
        if self.config_file is not None:
            logger.info(
                f"Config file already set. Replaced by the new file {file_path}."
            )
        self.config_file = file_path
        return {}

    def handle_eln_file(self, file_path: str) -> Dict[str, Any]:
        self.eln_data = parse_yml(
            file_path,
            convert_dict=CONVERT_DICT,
            parent_key="/ENTRY[entry]",
        )

        return {}

    def handle_rod_file(self, filepath) -> Dict[str, Any]:
        # specify default config file for rod files
        reader_dir = Path(__file__).parent
        self.config_file = reader_dir.joinpath("config", "config_file_rod.json")  # pylint: disable=invalid-type-comment

        rod = RodParser()
        # read the rod file
        rod.get_cif_file_content(filepath)
        # get the key and value pairs from the rod file
        self.raman_data = rod.extract_keys_and_values_from_cif()

        if self.raman_data.get("_raman_theoretical_spectrum.intensity"):
            logger.warning(
                f"Theoretical Raman Data .rod file found. File parsing aborted."
            )
            # prevent file parsing to setting an invalid config file name.
            self.config_file = Path()

        # unit_cell_alphabetagamma
        # replace the [ and ] to avoid confliucts in processing with pynxtools NXclass assignments
        self.raman_data = {
            key.replace("_[local]_", "_local_"): value
            for key, value in self.raman_data.items()
        }

        self.missing_meta_data = copy.deepcopy(self.raman_data)

        if self.raman_data.get("_cod_database_code") is not None or "":
            self.raman_data["COD_service_name"] = "Crystallography Open Database"
            del self.missing_meta_data["_cod_database_code"]

        if self.raman_data.get("_cell_length_a") is not None or "":
            # transform 9.40(3) to 9.40
            length_a = re.sub(r"\(\d+\)", "", self.raman_data.get("_cell_length_a"))
            length_b = re.sub(r"\(\d+\)", "", self.raman_data.get("_cell_length_b"))
            length_c = re.sub(r"\(\d+\)", "", self.raman_data.get("_cell_length_c"))
            self.raman_data["rod_unit_cell_length_abc"] = [
                float(length_a),
                float(length_b),
                float(length_c),
            ]
            del self.missing_meta_data["_cell_length_a"]
            del self.missing_meta_data["_cell_length_b"]
            del self.missing_meta_data["_cell_length_c"]
        if self.raman_data.get("_cell_angle_alpha") is not None or "":
            # transform 9.40(3) to 9.40
            angle_alpha = re.sub(
                r"\(\d+\)", "", self.raman_data.get("_cell_angle_alpha")
            )
            angle_beta = re.sub(r"\(\d+\)", "", self.raman_data.get("_cell_angle_beta"))
            angle_gamma = re.sub(
                r"\(\d+\)", "", self.raman_data.get("_cell_angle_gamma")
            )
            self.raman_data["rod_unit_cell_angles_alphabetagamma"] = [
                float(angle_alpha),
                float(angle_beta),
                float(angle_gamma),
            ]
            del self.missing_meta_data["_cell_angle_alpha"]
            del self.missing_meta_data["_cell_angle_beta"]
            del self.missing_meta_data["_cell_angle_gamma"]

        # This changes all uppercase string elements to lowercase string elements for the given key, within a given key value pair
        key_to_make_value_lower_case = "_raman_measurement.environment"
        environment_name_str = self.raman_data.get(key_to_make_value_lower_case)
        if environment_name_str is not None:
            self.raman_data[key_to_make_value_lower_case] = environment_name_str.lower()

        # transform the string into a datetime object
        time_key = "_raman_measurement.datetime_initiated"
        date_time_str = self.raman_data.get(time_key)
        if date_time_str is not None:
            date_time_obj = datetime.datetime.strptime(date_time_str, "%Y-%m-%d")
            # assume UTC for .rod data, as this is not specified in detail
            tzinfo = datetime.timezone.utc
            if isinstance(date_time_obj, datetime.datetime):
                if tzinfo is not None:
                    # Apply the specified timezone to the datetime object
                    date_time_obj = date_time_obj.replace(tzinfo=tzinfo)

                # assign the dictionary the corrrected date format
                self.raman_data[time_key] = date_time_obj.isoformat()

        # remove capitalization
        objective_type_key = "_raman_measurement_device.optics_type"
        objective_type_str = self.raman_data.get(objective_type_key)
        if objective_type_str is not None:
            self.raman_data[objective_type_key] = objective_type_str.lower()
            # set a valid raman NXDL value, but only if it matches one of the correct ones:
            objective_type_list = ["objective", "lens", "glass fiber", "none"]
            if self.raman_data.get(objective_type_key) not in objective_type_list:
                self.raman_data[objective_type_key] = "other"

        self.post_process = post_process_rod.__get__(self, RamanReader)

        return {}

    def handle_txt_file(self, filepath):
        """
        Read a .txt file from Witec Alpha Raman spectrometer and save the header and measurement data.
        """

        # specify default config file
        reader_dir = Path(__file__).parent
        self.config_file = reader_dir.joinpath("config", "config_file_witec.json")  # pylint: disable=invalid-type-comment

        self.raman_data = parse_txt_file(self, filepath)
        self.post_process = post_process_witec.__get__(self, RamanReader)

        return {}

    def get_eln_data(self, key: str, path: str) -> Any:
        """
        Returns data from the eln file. This is done via the file: "config_file.json".
        There are two sitations:
            1. The .json file has only a key assigned
            2. The .json file has a key AND a value assigned.
        The assigned value should be a "path", which reflects another entry in the eln file.
        This acts as eln_path redirection, which is used for example to assign flexible
        parameters from the eln_file (units, axisnames, etc.)
        """
        if self.eln_data is None:
            return None

        # Use the path to get the eln_data (this refers to the 2. case)
        if len(path) > 0:
            return self.eln_data.get(path)

        # If no path is assigned, use directly the given key to extract
        # the eln data/value (this refers to the 1. case)

        # Filtering list, for NeXus concepts which use mixed notation of
        # upper and lowercase to ensure correct NXclass labeling.
        upper_and_lower_mixed_nexus_concepts = [
            "/detector_TYPE[",
            "/beam_TYPE[",
            "/source_TYPE[",
            "/polfilter_TYPE[",
            "/spectral_filter_TYPE[",
            "/temp_control_TYPE[",
            "/software_TYPE[",
            "/OPTICAL_LENS[",
            "/identifierNAME[",
        ]
        if self.eln_data.get(key) is None:
            # filter for mixed concept names
            for string in upper_and_lower_mixed_nexus_concepts:
                key = key.replace(string, "/[")
            # add only characters, if they are lower case and if they are not "[" or "]"
            result = "".join(
                [char for char in key if not (char.isupper() or char in "[]")]
            )
            # Filter as well for
            result = result.replace("entry", f"ENTRY[{self.callbacks.entry_name}]")

            if self.eln_data.get(result) is not None:
                return self.eln_data.get(result)
            else:
                logger.warning(
                    f"No key found during eln_data processsing for key '{key}' after it's modification to '{result}'."
                )
        return self.eln_data.get(key)

    def get_data(self, key: str, path: str) -> Any:
        """
        Returns the data from a .rod file (Raman Open Database), which was trasnferred into a dictionary.
        """

        value = self.raman_data.get(path)

        # this filters out the meta data, which is up to now only created for .rod files

        if (path is None or path == "") and key is not None:
            return self.raman_data.get(key)

        if self.missing_meta_data:
            # this if condition is required, to only delete keys which are abaialble by the data.
            # e.g. is defined to extract it via config.json, but there is no value in meta data
            if path in self.missing_meta_data.keys():
                del self.missing_meta_data[path]

        if value is not None:
            try:
                # ensure that the space_group entry from NXsample is of type
                # NXchar, even if space group numbers are used
                if "/space_group" in key and "/SAMPLE" in key:
                    return value
                return float(value)
            except (ValueError, TypeError):
                return self.raman_data.get(path)
        else:
            logger.warning(f"No axis name corresponding to the path {path}.")

    def read(
        self,
        template: dict = None,
        file_paths: Tuple[str] = None,
        objects: Tuple[Any] = None,
        **kwargs,
    ) -> dict:
        template = super().read(template, file_paths, objects, suppress_warning=True)
        # set default data

        if self.missing_meta_data:
            for key in self.missing_meta_data:
                template[
                    f"/ENTRY[{self.callbacks.entry_name}]/COLLECTION[unused_rod_keys]/{key}"
                ] = f"{self.missing_meta_data[key]}"

        template["/@default"] = "entry"

        return template


READER = RamanReader



import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Union
import numpy as np


logger = logging.getLogger("pynxtools")


def parse_txt_file(self, filepath):
    """
    Read a .txt file from Witec Alpha Raman spectrometer and return a data dictionary
    which contains Raman shift and Intensity
    """
    with open(filepath, "r") as file:
        lines = file.readlines()

    # Initialize dictionaries to hold header and data sections
    header_dict = {}
    data = []
    line_count = 0
    data_mini_header_length = None

    # Track current section
    current_section = None

    for line in lines:
        line_count += 1
        # Remove any leading/trailing whitespace
        line = line.strip()
        # Go through the lines and define two different regions "Header" and
        # "Data", as these need different methods to extract the data.
        if line.startswith("[Header]"):
            current_section = "header"
            continue
        elif line.startswith("[Data]"):
            data_mini_header_length = line_count + 2
            current_section = "data"

            continue

        # Parse the header section
        if current_section == "header" and "=" in line:
            key, value = line.split("=", 1)
            header_dict[key.strip()] = value.strip()

        # Parse the data section
        elif current_section == "data" and "," in line:
            # The header is set excactly until the float-like column data starts
            # Rework this later to extract full metadata
            if line_count <= data_mini_header_length:
                if line.startswith("[Header]"):
                    logger.info(
                        f"[Header] elements in the file {filepath}, are not parsed yet. Consider adden the respective functionality."
                    )
            if line_count > data_mini_header_length:
                values = line.split(",")
                data.append([float(values[0].strip()), float(values[1].strip())])

    # Transform: [[A, B], [C, D], [E, F]] into [[A, C, E], [B, D, F]]
    data = [list(item) for item in zip(*data)]

    # assign column data with keys
    data_dict = {"data/x_values": data[0], "data/y_values": data[1]}
    return data_dict


def post_process_witec(self) -> None:
    """
    Post process the Raman data to add the Raman Shift from input laser wavelength and
    data wavelengths.
    """

    def transform_nm_to_wavenumber(lambda_laser, lambda_measurement):
        stokes_raman_shift = -(
            1e7 / np.array(lambda_measurement) - 1e7 / np.array(lambda_laser)
        )
        # return a list as output
        return stokes_raman_shift.tolist()

    def get_incident_wavelength_from_NXraman():
        substring = "/beam_incident/wavelength"

        # Find matching keys with contain this substring
        wavelength_keys = [key for key in self.eln_data if substring in key]
        # Filter the matching keys for the strings, which contain this substring at the end only
        filtered_list = [
            string for string in wavelength_keys if string.endswith(substring)
        ]
        # get the laser wavelength
        laser_wavelength = self.eln_data.get(filtered_list[0])
        return laser_wavelength

    laser_wavelength = get_incident_wavelength_from_NXraman()

    x_values_raman = transform_nm_to_wavenumber(
        laser_wavelength, self.raman_data["data/x_values"]
    )

    # update the data dictionary
    self.raman_data["data/x_values_raman"] = x_values_raman



import logging
from pathlib import Path
from typing import Any, Dict, List, Union

import gemmi  # for cif file handling
import numpy as np

logger = logging.getLogger("pynxtools")


class RodParser:
    """
    This class provides the ultilieies to read in a .rod file with "get_cif_file_content".
    Then extract all data via "extract_keys_and_values_from_cif" into a dictionary.
    """

    def __init__(self, *args, **kwargs):
        self.cif_doc = None
        self.cif_block = None
        self.lines = []

    def _read_lines(self, file: Union[str, Path]):
        """
        Read all lines from the input files.
        """
        with open(file, encoding="utf-8") as utf8_file:
            lines = utf8_file.readlines()

        return lines

    def get_cif_file_content(self, file_path):
        doc = gemmi.cif.read_file(file_path)
        block = doc.sole_block()  # extract main block of cif file
        self.cif_doc = doc
        self.cif_block = block
        self.lines = self._read_lines(file_path)

    def get_string_position(self, string_element: str, check_only_pos_zero=False):
        line_positions_of_str_element = []

        rod_lines = self.lines
        if check_only_pos_zero:
            for line_number, lines in enumerate(rod_lines):
                if string_element in lines[0]:
                    line_positions_of_str_element.append(line_number)
        else:
            if rod_lines is None:
                logger.info(f"Problem during reading .rod file. 'rod_line' is None.")
            else:
                for line_number, lines in enumerate(rod_lines):
                    if string_element in lines:
                        line_positions_of_str_element.append(line_number)
        return line_positions_of_str_element

    def get_keys_and_loop_boolean(self, key_positions, key_pos_in_loops):
        cif_key_loop_boolean_dict = {}
        for key_pos in key_positions:
            # go through all key_positions (i.e. line number)
            # and check two cases: These lines are part of a loop or
            # they are not part of a loop
            # If they are in a loop, assign the bool value required for read
            # out of the value from the key (i.e set =True)
            if key_pos in key_pos_in_loops:
                # remove linebreaks to ensure right assignment in values for input keys
                cif_key_loop_boolean_dict[self.lines[key_pos].replace("\n", "")] = True
            if key_pos not in key_pos_in_loops:
                # some keys have their values on the same line, some on other lines
                # Extract only the key, as this is always avaialble.
                # Use the key later to get the respective values
                if " " in self.lines[key_pos]:
                    key, value = self.lines[key_pos].split(maxsplit=1)
                    cif_key_loop_boolean_dict[key] = False
                else:
                    # If only the key is on the line, without its value, extract only the key,
                    # Remove possible linebreaks for clarity with .replace()
                    cif_key_loop_boolean_dict[self.lines[key_pos].replace("\n", "")] = (
                        False
                    )

        if len(key_positions) == len(cif_key_loop_boolean_dict):
            return cif_key_loop_boolean_dict
        else:
            logger.info(f".rod file parsing warning: Not all rod-keys were parsed.")
            return cif_key_loop_boolean_dict

    def key_pos_after_loop(self, loop_pos_lists, key_pos_list):
        loop_key_positons = []
        for loop_pos_list in loop_pos_lists:
            counter = 1
            while loop_pos_list + counter in key_pos_list:
                if (
                    counter >= 100
                ):  # implemented to avoid infinite loop, how to do better?
                    raise IndexError
                loop_key_positons.append(loop_pos_list + counter)
                counter += 1

        return loop_key_positons

    def get_cif_value_from_key(
        self, value_key: str, is_cif_loop_value=False
    ) -> Union[str, list]:
        """
        Parse the top-level Prodigy export settings into a dict.

        Parameters
        ----------
        value_key : str
            name of the key value, which is used for extraction

        is_cif_loop_value : boolean
            if the key value, is part of a loop structure, this has to be set
            correctly to extracat the respective array-like values

        Returns
        -------
        output_list : str, list or np.array
            Values, list, or np.array which is assigned to the respective key in the cif file

        """

        block = self.cif_block  # extract main block of cif file
        if not is_cif_loop_value:  # is single value via _key = value
            value = block.find_value(value_key)
            # perform processing if string is not single line value
            if value.count("\n") > 0:
                value = value.replace(";\n", "")
                value = value.replace("\n;", "")
                if value.count("\n") > 0:
                    value = value.replace("\n", " ")
                return value.lstrip()  # remove leading space if it is present
            if value.count("\n") == 0:
                if value.startswith("'"):
                    return value.replace("'", "")
                return value
        if is_cif_loop_value:  # if block like value via loop_ = [....]
            output_list = []
            for element in block.find_loop(value_key):
                output_list.append(element)
            # try: # try to conver tto numpy array
            #    output_list = np.array(output_list, dtype=float)
            #    return output_list
            try:  # try to conver tto numpy array
                output_list_float = [float(item) for item in output_list]
                return output_list_float
            except ValueError:  # default string output if not convertable to float
                return output_list
        return None

    def extract_keys_and_values_from_cif(self):
        loop_positions = self.get_string_position("loop_\n")
        key_pos_non_loop = self.get_string_position("_", check_only_pos_zero=True)
        key_pos_in_loops = self.key_pos_after_loop(loop_positions, key_pos_non_loop)
        cif_key_dict_with_loop_boolean = self.get_keys_and_loop_boolean(
            key_pos_non_loop, key_pos_in_loops
        )

        # create a dictionary, and extract all the values by using the keys in correct formatting
        cif_dict_key_value_pair_dict = {}
        for key in cif_key_dict_with_loop_boolean:
            bool_loop_value = cif_key_dict_with_loop_boolean[key]
            cif_dict_key_value_pair_dict[key] = self.get_cif_value_from_key(
                key, is_cif_loop_value=bool_loop_value
            )

        return cif_dict_key_value_pair_dict


def post_process_rod(self) -> None:
    wavelength_nm = float(
        self.raman_data.get("_raman_measurement_device.excitation_laser_wavelength")
    )
    resolution_invers_cm = float(
        self.raman_data.get("_raman_measurement_device.resolution")
    )

    if wavelength_nm is not None and resolution_invers_cm is not None:
        # assume the resolution is referd to the resolution at the laser wavelength
        wavelength_invers_cm = 1e7 / wavelength_nm
        resolution_nm = resolution_invers_cm / wavelength_invers_cm * wavelength_nm

        # update the data dictionary
        self.raman_data[
            "/ENTRY[entry]/INSTRUMENT[instrument]/wavelength_resolution/physical_quantity"
        ] = "wavelength"
        self.raman_data[
            "/ENTRY[entry]/INSTRUMENT[instrument]/wavelength_resolution/resolution"
        ] = resolution_nm
        self.raman_data[
            "/ENTRY[entry]/INSTRUMENT[instrument]/wavelength_resolution/resolution/@units"
        ] = "nm"
        # remove this key from original input data
        del self.missing_meta_data["_raman_measurement_device.resolution"]

    diffraction_grating = self.raman_data.get(
        "_raman_measurement_device.diffraction_grating"
    )

    if diffraction_grating is not None:
        self.raman_data[
            "/ENTRY[entry]/INSTRUMENT[instrument]/MONOCHROMATOR[monochromator]/GRATING[grating]/period"
        ] = 1 / float(diffraction_grating)



try:
    from nomad.config.models.plugins import AppEntryPoint
except ImportError as exc:
    raise ImportError(
        "Could not import nomad package. Please install the package 'nomad-lab'."
    ) from exc

from nomad.config.models.ui import (
    App,
    Column,
    Menu,
    MenuItemHistogram,
    MenuItemPeriodicTable,
    MenuItemTerms,
    MenuSizeEnum,
    SearchQuantities,
)

schema = "pynxtools.nomad.schema.Root"

raman_app = AppEntryPoint(
    name="RamanApp",
    description="Simple Raman app.",
    app=App(
        # Label of the App
        label="Raman",
        # Path used in the URL, must be unique
        path="ramanapp",
        # Used to categorize apps in the explore menu
        category="Experiment",
        # Brief description used in the app menu
        description="A simple search app customized for Raman data.",
        # Longer description that can also use markdown
        readme="This is a simple App to support basic search for Raman based Experiment Entries.",
        # If you want to use quantities from a custom schema, you need to load
        # the search quantities from it first here. Note that you can use a glob
        # syntax to load the entire package, or just a single schema from a
        # package.
        search_quantities=SearchQuantities(
            include=[f"*#{schema}"],
            # include=[f"data.Raman.*#{schema}"],
        ),
        # Controls which columns are shown in the results table
        columns=[
            Column(quantity="entry_id", selected=True),
            Column(
                title="Material Name",
                quantity=f"data.ENTRY[*].SAMPLE[*].name__field#{schema}",
                selected=True,
            ),
            Column(
                title="Space Group Number",
                quantity=f"data.ENTRY[*].SAMPLE[*].space_group__field#{schema}#str",
                selected=True,
            ),
            # Column(
            #    title="Temperature",
            #    quantity=f"data.ENTRY[*].SAMPLE[*].ENVIRONMENT[1].SENSOR[*].value__field#{schema}",
            #    selected=True,
            # ),
            Column(
                title="Unit Cell Volume",
                quantity=f"data.ENTRY[*].SAMPLE[*].unit_cell_volume__field#{schema}",
                selected=True,
            ),
            Column(
                title="Long Name",
                quantity=f"data.ENTRY[*].title__field#{schema}",
                selected=True,
            ),
        ],
        # Dictionary of search filters that are always enabled for queries made
        # within this app. This is especially important to narrow down the
        # results to the wanted subset. Any available search filter can be
        # targeted here. This example makes sure that only entries that use
        # MySchema are included.
        # filters_locked={"section_defs.definition_qualified_name": [schema]},
        filters_locked={f"data.ENTRY.definition__field#{schema}": ["NXraman"]},
        # Controls the menu shown on the left
        menu=Menu(
            title="Material",
            items=[
                Menu(
                    title="Elements",
                    size=MenuSizeEnum.XXL,
                    items=[
                        MenuItemPeriodicTable(
                            search_quantity="results.material.elements",
                        ),
                        MenuItemTerms(
                            search_quantity="results.material.chemical_formula_hill",
                            width=6,
                            options=0,
                        ),
                        MenuItemTerms(
                            search_quantity="results.material.chemical_formula_iupac",
                            width=6,
                            options=0,
                        ),
                        MenuItemTerms(
                            search_quantity="results.material.chemical_formula_reduced",
                            width=6,
                            options=0,
                        ),
                        MenuItemTerms(
                            search_quantity="results.material.chemical_formula_anonymous",
                            width=6,
                            options=0,
                        ),
                        MenuItemHistogram(
                            x="results.material.n_elements",
                        ),
                    ],
                ),
                Menu(
                    title="Space Group Number",
                    items=[
                        MenuItemTerms(
                            quantity=f"data.ENTRY.SAMPLE.space_group__field#{schema}#str",
                            width=10,
                            options=10,
                        ),
                    ],
                ),
                Menu(
                    title="Raman Spectrometer Model",
                    items=[
                        MenuItemTerms(
                            quantity=f"data.ENTRY.INSTRUMENT.device_information.model__field#{schema}#str",
                            width=10,
                            options=5,
                        ),
                    ],
                ),
                Menu(
                    title="Scattering Configuration",
                    items=[
                        MenuItemTerms(
                            quantity=f"data.ENTRY.INSTRUMENT.scattering_configuration__field#{schema}#str",
                            width=10,
                            options=7,
                        ),
                    ],
                ),
                Menu(
                    title="Instruments",
                    size=MenuSizeEnum.LG,
                    items=[
                        MenuItemTerms(
                            title="Name",
                            search_quantity=f"data.ENTRY.INSTRUMENT.name__field#{schema}",
                            width=12,
                            options=12,
                        ),
                        MenuItemTerms(
                            title="Short Name",
                            search_quantity=f"data.ENTRY.INSTRUMENT.name___short_name#{schema}",
                            width=12,
                            options=12,
                        ),
                    ],
                ),
                Menu(
                    title="Samples",
                    size=MenuSizeEnum.LG,
                    items=[
                        MenuItemTerms(
                            title="Name",
                            search_quantity=f"data.ENTRY.SAMPLE.name__field#{schema}",
                            width=12,
                            options=12,
                        ),
                        MenuItemTerms(
                            title="Sample ID",
                            search_quantity=f"data.ENTRY.SAMPLE.identifierNAME__field#{schema}",
                            width=12,
                            options=12,
                        ),
                    ],
                ),
                Menu(
                    title="Authors / Origin",
                    size=MenuSizeEnum.LG,
                    items=[
                        MenuItemTerms(
                            title="Entry Author",
                            search_quantity=f"data.ENTRY.USER.name__field#{schema}",
                            width=12,
                            options=5,
                        ),
                        MenuItemTerms(
                            title="Upload Author",
                            search_quantity=f"authors.name",
                            width=12,
                            options=5,
                        ),
                        MenuItemTerms(
                            title="Affiliation",
                            search_quantity=f"data.ENTRY.USER.affiliation__field#{schema}",
                            width=12,
                            options=5,
                        ),
                    ],
                ),
                MenuItemHistogram(
                    title="Start Time",
                    x=f"data.ENTRY.start_time__field#{schema}",
                    autorange=True,
                ),
                MenuItemHistogram(
                    title="Upload Creation Time",
                    x=f"upload_create_time",
                    autorange=True,
                ),
            ],
        ),
        # Controls the default dashboard shown in the search interface
        dashboard={
            "widgets": [
                {
                    "type": "histogram",
                    "show_input": False,
                    "autorange": True,
                    "nbins": 30,
                    "scale": "log",
                    "quantity": f"data.ENTRY.INSTRUMENT.beam_incident.wavelength__field#{schema}#float",
                    "title": "Incident Wavelength [nm]",
                    "layout": {
                        "lg": {"minH": 3, "minW": 3, "h": 5, "w": 8, "y": 0, "x": 0}
                    },
                },
                {
                    "type": "histogram",
                    "show_input": False,
                    "autorange": True,
                    "nbins": 30,
                    "scale": "log",
                    "quantity": f"data.ENTRY.INSTRUMENT.beam_incident.average_power__field#{schema}#float",
                    "title": "Laser Power [mW]",
                    "layout": {
                        "lg": {"minH": 3, "minW": 3, "h": 4, "w": 8, "y": 5, "x": 0}
                    },
                },
                {
                    "type": "histogram",
                    "show_input": False,
                    "autorange": True,
                    "nbins": 30,
                    "scale": "log",
                    "quantity": f"data.ENTRY.INSTRUMENT.LENS_OPT.magnification__field#{schema}#float",
                    "title": "Magnification",
                    "layout": {
                        "lg": {"minH": 3, "minW": 3, "h": 3, "w": 6, "y": 0, "x": 8}
                    },
                },
                {
                    "type": "histogram",
                    "show_input": False,
                    "autorange": True,
                    "nbins": 30,
                    "scale": "log",
                    "quantity": f"data.ENTRY.INSTRUMENT.LENS_OPT.numerical_aperture__field#{schema}#float",
                    "title": "Numerical Aperture",
                    "layout": {
                        "lg": {"minH": 3, "minW": 3, "h": 3, "w": 6, "y": 3, "x": 8}
                    },
                },
                {
                    "type": "histogram",
                    "show_input": False,
                    "autorange": True,
                    "nbins": 30,
                    "scale": "log",
                    "quantity": f"data.ENTRY.INSTRUMENT.beam_incident.extent__field#{schema}#float",
                    "title": "Beam diameter [µm]",
                    "layout": {
                        "lg": {"minH": 3, "minW": 3, "h": 3, "w": 6, "y": 6, "x": 8}
                    },
                },
            ]
        },
    ),
)
